#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <pthread.h>
#include <infiniband/verbs.h>
#include <infiniband/verbs_exp.h>
#include <sys/types.h>
#include "common.h"

void *connection_handler(void *);

struct test_ctx *global_ctx;
pthread_t sniffer_thread;

static void sig_catcher(int signo) {
    if (ibv_exp_destroy_dct(global_ctx->dct)) {
        fprintf(stderr, "failed to destroy dct\n");
        exit(1);
    }
    
    /* TODO: clean all ib resources.. */
    exit(1);
}

int create_recv_wqes(struct test_ctx *ctx)
{
    struct ibv_recv_wr      *bad_wr_recv;
    int 			j;
    
    ctx->recv_sge_list.addr  = (uintptr_t)ctx->addr;
    ctx->recv_sge_list.length = ctx->size;
    ctx->recv_sge_list.lkey   = ctx->mr->lkey;
    
    ctx->rwr.sg_list	= &ctx->recv_sge_list;
    ctx->rwr.wr_id		= 0;
    ctx->rwr.next		= NULL;
    ctx->rwr.num_sge	= 1;
    
    ctx->rx_buffer_addr = ctx->recv_sge_list.addr;
    
    for (j = 0; j < ctx->rx_depth ; ++j) {
        
        if (ibv_post_srq_recv(ctx->srq, &ctx->rwr, &bad_wr_recv)) {
            fprintf(stderr, "Couldn't post recv SRQ");
            return 1;
        }
        
        if (ctx->size <= (ctx->page_size / 2)) {
            increase_loc_addr(&ctx->recv_sge_list,
                              ctx->size,
                              j,
                              ctx->rx_buffer_addr,
                              0,
                              ctx->cache_line_size,
                              ctx->page_size);
        }
    }
    
    ctx->recv_sge_list.addr = ctx->rx_buffer_addr;
    
    return 0;
}

int init_ib_resources(struct test_ctx *ctx)
{
    ctx->dev_list = ibv_get_device_list(NULL);
    
    if (!ctx->dev_list) {
        perror("Failed to get IB devices list");
        return 1;
    }
    
    if (!ctx->ib_devname) {
        ctx->ib_dev = *ctx->dev_list;
        if (!ctx->ib_dev) {
            fprintf(stderr, "No IB devices found\n");
            return 1;
        }
    } else {
        int i; for (i = 0; ctx->dev_list[i]; ++i) {
            if (strcmp(ibv_get_device_name(ctx->dev_list[i]), ctx->ib_devname)) continue;
            ctx->ib_dev = ctx->dev_list[i];
        }
        if (!ctx->ib_dev) {
            fprintf(stderr, "IB device %s not found\n", ctx->ib_devname);
            return 1;
        }
    }
    
    ctx->ctx = ibv_open_device(ctx->ib_dev);
    if (!ctx->ctx) {
        fprintf(stderr, "Couldn't get context for %s\n",
                ibv_get_device_name(ctx->ib_dev));
        return 1;
    }
    
    ctx->device_name_by_id = ib_dev_name(ctx->ctx);
    if (ctx->device_name_by_id == DEVICE_ERROR) {
        fprintf(stderr, "Couldn't get device name\n");
        return 1;
    }
    
    /* TODO: do query device instead of this */
    switch (ctx->device_name_by_id)
    {
        case CONNECTIB:
        case CONNECTX4:
        case CONNECTX4LX:
        case CONNECTX5:
        case CONNECTX5EX:
            break;
        default:
            fprintf(stderr, "DC transport is not suppported for this device\n");
            return 1;
    }
    
    ctx->pd = ibv_alloc_pd(ctx->ctx);
    if (!ctx->pd) {
        fprintf(stderr, "failed to allocate pd\n");
        return 1;
    }
    
    if (create_mr(ctx))
    {
        fprintf(stderr,"failed to create mr\n");
        return 1;
    }
    
    ctx->cq = ibv_create_cq(ctx->ctx, ctx->rx_depth*ctx->num_of_qps, NULL, NULL, 0);
    if (!ctx->cq) {
        fprintf(stderr, "failed to create cq\n");
        return 1;
    }
    
    struct ibv_srq_init_attr attr = {
        .attr = {
            .max_wr  = ctx->rx_depth,
            .max_sge = 1
        }
    };
    
    ctx->srq = ibv_create_srq(ctx->pd, &attr);
    if (!ctx->srq)  {
        fprintf(stderr, "Couldn't create SRQ\n");
        return 1;
    }
    
    ibv_get_srq_num(ctx->srq, &ctx->srqn);
    
    if (ibv_query_port(ctx->ctx, ctx->ib_port, &ctx->portinfo)) {
        fprintf(stderr, "Couldn't get port info\n");
        return 1;
    }
    if (ctx->user_mtu) {
        if (ctx->mtu > ctx->portinfo.active_mtu) {
            fprintf(stderr, "Requested MTU (%d) larget than port MTU (%d)\n", ctx->user_mtu, ctx->portinfo.active_mtu);
            return 1;
        }
    } else ctx->mtu = ctx->portinfo.active_mtu;
    
    if (ctx->verb == SEND) {
        if (create_recv_wqes(ctx)) {
            fprintf(stderr, "Couldn't create recv wqes\n");
            return 1;
        }
    }
    if (ibv_query_gid(ctx->ctx, ctx->ib_port, ctx->gid_index, &ctx->gid)) return 1;
    
    return 0;
}

int create_dct(struct test_ctx *ctx)
{
    struct ibv_exp_dct_init_attr dctattr = {
        .pd = ctx->pd,
        .cq = ctx->cq,
        .srq = ctx->srq,
        .dc_key = ctx->dct_key,
        .port = ctx->ib_port,
        .access_flags = IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE,
        .min_rnr_timer = 2,
        .tclass = ctx->tclass,
        .flow_label = 0,
        .mtu = ctx->mtu,
        .pkey_index = 0,
        .gid_index = ctx->gid_index,
        .hop_limit = 1,
        .create_flags = 0,
        .inline_size = ctx->inl,
    };
    
    ctx->dct = ibv_exp_create_dct(ctx->ctx, &dctattr);
    if (!ctx->dct) {
        printf("create dct failed\n");
        return -1;
    }
    printf("created DCT=0x%.6x\n", ctx->dct->dct_num);

    return 1;
}

void send_data_to_client(int sock, test_data data)
{
    
    char msg[MSG_FORMAT_SIZE];
    sprintf(msg, MSG_FORMAT, data.dct_num, data.dct_key, data.lid, data.gid.global.subnet_prefix, data.gid.global.interface_id, data.psn, data.rkey, data.vaddr, data.srqn, data.gid_index, data.mtu);
    
    if (write(sock, msg, sizeof(msg)) != sizeof(msg)) {
        fprintf(stderr, "Couldn't send local address\n");
    }
}

void *connection_handler(void *h_data)
{
    handler_data l_data = *((handler_data*)h_data);
    int sock = *(l_data.socket);
    
    send_data_to_client(sock, l_data.data);
    free(l_data.socket);
    free(h_data);
    return 0;
}

//int start_connection_manager(int port, test_data *data)
void *start_connection_manager(void *h_data)
{
    int port = ((handler_data*)h_data)->tcp_port;
    
    test_data *data = &((handler_data*)h_data)->data;
    char *logname = ((handler_data*)h_data)->logname;
    
    int socket_desc , client_sock , c , *new_sock;
    struct sockaddr_in server , client;
    
    socket_desc = socket(AF_INET , SOCK_STREAM , 0);
    if (socket_desc == -1)
    {
        printf("Could not create socket\n");
    }
    
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons( port );
    if (bind(socket_desc,(struct sockaddr *)&server , sizeof(server)) < 0)
    {
        perror("bind failed. Error");
        exit(1);
    }
    
    listen(socket_desc , 150);
    
    puts("Waiting for incoming connections...");
    c = sizeof(struct sockaddr_in);
    
    while ((client_sock = accept(socket_desc, (struct sockaddr *)&client, (socklen_t*)&c)))
    {
        pthread_t sniffer_thread;
        new_sock = malloc(sizeof(client_sock));
        *new_sock = client_sock;

        printf("New connection from client IP: %s, listen_port: %d, port: %d\n", inet_ntoa(client.sin_addr), port, ntohs(client.sin_port));

        FILE* file;
        if (logname) {
        	file = fopen(logname, "a"); // write to the end of the file.
        	fprintf(file, "New connection from client IP: %s, listen_port: %d, port: %d\n", inet_ntoa(client.sin_addr), port, ntohs(client.sin_port));
        	fclose(file);
        }
        
        //init handler data
        handler_data *h_data = malloc(sizeof(handler_data));
        copy_test_data(data, &h_data->data);
        h_data->socket = new_sock;
        h_data->tcp_port = port;
        
        if (pthread_create( &sniffer_thread , NULL ,  connection_handler , (void*) h_data) < 0)
        {
            perror("could not create thread");
            exit(1);
        }
    }
    if (client_sock < 0)
    {
        perror("accept failed");
        exit(1);
    }
    
    return NULL;
}


void poll_cq(struct test_ctx *ctx, test_data *data)
{
    struct ibv_wc           *wc          = NULL;
    struct ibv_recv_wr      *bad_wr_recv = NULL;
    int			poll_batch   = 16;
    int			ne, i;
    int			rcnt = 0;
    
    ALLOCATE(wc, struct ibv_wc, poll_batch);
    
    while (1)
    {
        ne = ibv_poll_cq(ctx->cq, poll_batch, wc);
        
        if (ne > 0)
        {
            for(i = 0; i < ne; i++)
            {
                if (wc[i].status != IBV_WC_SUCCESS)
                {
                    fprintf(stderr, "got CQE with error.. exiting\n");
                    exit(1);
                }
                
                if (ibv_post_srq_recv(ctx->srq, &ctx->rwr, &bad_wr_recv))
                {
                    fprintf(stderr, "failed to post_srq_recv\n");
                    exit(1);
                }
                rcnt++;
                if (ctx->size <= (ctx->page_size / 2))
                    increase_loc_addr(ctx->rwr.sg_list, ctx->size, rcnt, ctx->rx_buffer_addr, 0, ctx->cache_line_size, ctx->page_size);
            }
        }
        else if (ne < 0)
        {
            fprintf(stderr, "error in poll cq\n");
            exit(1);
        }
    }
}

int init_connection_manager_thread(struct test_ctx *ctx, test_data *data, handler_data *h_data)
{
    //init handler data
    copy_test_data(data, &h_data->data);
    h_data->tcp_port = ctx->port;
    
    if (pthread_create( &sniffer_thread, NULL, start_connection_manager, (void*) h_data) < 0)
    {
        perror("could not create thread");
        return -1;
    }
    
    return 0;
}

void force_sig_catcher_on_parent(sigset_t *set)
{
    /* disable signal catcher on child threads */
    sigemptyset(set);
    sigaddset (set, SIGINT);
    pthread_sigmask(SIG_SETMASK, set, NULL);
}

void init_sig_catcher(struct test_ctx *ctx)
{
    global_ctx = ctx;
    if (signal(SIGINT, sig_catcher) == SIG_ERR) {
        printf("failed to init signal handler\n");
        exit(1);
    }
}

char* create_logname_server(char* file_path, char* port_str) {
	char* name = "/dc_server_logfile_port_";
	char *result = malloc(strlen(file_path) + strlen(name) + strlen(port_str)+1);//+1 for the null-terminator
	strcpy(result, file_path);
	strcat(result, name);
	strcat(result, port_str);
	return result;
}

int main(int argc , char *argv[])
{
    struct test_ctx	ctx;
    test_data	data;
    sigset_t set;
    handler_data h_data;
    
    ibv_fork_init();
    
    init_ctx(&ctx, argc, argv, 0);
    char **ports_array = (char**)malloc((ctx.n_ports+1)*sizeof(*ports_array));
    if (parse_port_list(&ctx, &ports_array)){
        destroy_arrays(NULL, ports_array);
        exit(1);
    }
        
    if (1<ctx.n_ports) {
        fprintf(stderr, "Error, server only supports a single listen port (currently)\n");
        exit(1);
    }
    
    FILE* file;
    if (NULL == ctx.file_path) {
    	h_data.logname = NULL;
    } else { // there is a logfile
    	h_data.logname = create_logname_server(ctx.file_path, ctx.port_str);
    	printf("logname: %s\n", h_data.logname);
    	file = fopen(h_data.logname, "w");
		if (NULL == file) {
			fprintf(stderr, "Error opening file! Please check the file_path you gave (-F)\n");
			exit(1);
		}
    }
    
    print_command(h_data.logname, argc, argv);

    init_sig_catcher(&ctx);
    
    if (init_ib_resources(&ctx)) {
        fprintf(stderr, "Failed to create IB resources.\n");
        return 1;
    }
    
    if (create_dct(&ctx) == -1) return 1; // FAIL
    
    init_test_data(&ctx, &data);
    
    if (init_connection_manager_thread(&ctx, &data, &h_data) == -1) return 1;  // FAIL

    force_sig_catcher_on_parent(&set);
    
    if (ctx.verb == SEND)
        poll_cq(&ctx, &data);
    
    pthread_join(sniffer_thread, NULL);
    destroy_arrays(NULL, ports_array);
    
    return 0;
}
